pub const SERVERLISTNAME: &str = "servers.list";

use std::{
    fmt::Display,
    fs::{self, File},
    io::{ErrorKind, Read},
    net::SocketAddr,
    str::FromStr,
};

lazy_static! {
    // Assuming the regex is valid in the source code, this should NEVER fail,
    // So I'm assuaged with the possible panic here.
    static ref SERVER_REGEX: Regex = Regex::new(r"(\w+)\(([^\)]+)\)")
        .wrap_err("While compiling regex")
        .expect("Unable to compile regex");
}

use color_eyre::Report;
use eyre::Context;
use regex::Regex;
use tracing::error;

#[derive(Debug)]
pub struct QueryServer {
    pub name: String,
    pub socket: SocketAddr,
    pub server_type: ServerType,
}

#[derive(Debug)]
pub enum ServerTypeParseError {
    Generic(String),
}

impl Display for ServerTypeParseError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let ServerTypeParseError::Generic(inner) = self;
        write!(f, "{}", inner)
    }
}

#[derive(Debug)]
pub enum ServerType {
    Source,
    Minecraft,
    Mumble,
}

impl<'a> TryFrom<&'a str> for ServerType {
    type Error = ServerTypeParseError;

    fn try_from(value: &'a str) -> Result<Self, Self::Error> {
        let value = value.to_lowercase();
        match value.as_str() {
            "source" => Ok(ServerType::Source),
            "mumble" => Ok(ServerType::Mumble),
            "minecraft" => Ok(ServerType::Minecraft),
            _ => Err(ServerTypeParseError::Generic(format!(
                "Unknown Value: {}",
                value
            ))),
        }
    }
}

impl std::fmt::Display for QueryServer {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{} ({})", self.name, self.socket)
    }
}

fn default_servers() -> String {
    String::from(
        "example-server\
        \nSource(127.0.0.1:27015)",
    )
}

fn parse_query_server(sv_name: &str, ln: &str) -> Result<QueryServer, eyre::Report> {
    let unmatch = "Regex unable to get match";
    let captures = SERVER_REGEX
        .captures(ln)
        .ok_or_else(|| eyre::Report::msg("Regex unable to match anything"))?;

    let server_type = ServerType::try_from(
        captures
            .get(1)
            .ok_or_else(|| eyre::Report::msg(unmatch))?
            .as_str(),
    )
    .map_err(|e| eyre::Report::msg(format!("Unable to get server type while parsing: {}", e)))?;
    let inner_value = captures
        .get(2)
        .ok_or_else(|| eyre::Report::msg(unmatch))?
        .as_str();

    let socket = SocketAddr::from_str(inner_value).wrap_err("When parsing address")?;

    Ok(QueryServer {
        name: sv_name.into(),
        socket,
        server_type,
    })
}

/// Convert the string into a vector of QueryServer structs.
///
/// This function never fails,
/// but may return an empty vector if unable to parse any valuable input.
fn parse_lines(list: String) -> Vec<QueryServer> {
    let mut vec = Vec::new();
    let mut lines = list.lines();
    let mut ln = 1;

    // Primary while loop iterates over every other line, as we iterate again,
    // inside a match block to parse the IP address.
    while let Some(item) = lines.next() {
        let name = item.trim();
        ln += 1;
        match lines.next() {
            Some(v) => match parse_query_server(name, v) {
                Ok(qs) => {
                    vec.push(qs);
                }
                Err(e) => {
                    let msg = format!(
                        "Error parsing server.list, at line: {}\n\
                            Value:\n    {}",
                        ln, v,
                    );
                    let report: Report = e;
                    error!("{:?}", report.wrap_err(msg))
                }
            },
            // Just log the error and continue on. Non-fatal.
            None => {
                let err = format!(
                    "Error when parsing server list, unexpected EoF at line: {}",
                    ln
                );
                error!("{}", err);
            }
        };
        ln += 1;
    }
    vec
}

/// Loads the servers from the `server.list` beside the exectuable.
///
/// Returns an error if the list file:
/// * Cannot be parsed
/// * Does not exist, and cannot be created.
/// * Cannot be accessed
pub fn get_servers() -> Result<Vec<QueryServer>, eyre::Report> {
    // This sets the configuration file name relative to the exe location.
    let mut path = std::env::current_exe()?;
    path.pop();
    path.push(SERVERLISTNAME);

    let servers = parse_lines(default_servers());

    // Check if the list file exists, and can be written to.
    match File::open(&path) {
        Ok(mut f) => {
            let mut buf = String::new();
            f.read_to_string(&mut buf)?;
            Ok(parse_lines(buf))
        }
        Err(e) => match e.kind() {
            ErrorKind::NotFound => {
                error!("No server list found...");
                fs::write(&path, default_servers())?;
                error!("Wrote example to '{}'", &path.to_string_lossy());
                Ok(servers)
            }
            _ => Err(e)
                .wrap_err_with(|| format!("Unable to access cfg file {}", &path.to_string_lossy())),
        },
    }
}
