pub const AUTHORS: &str = env!("CARGO_PKG_AUTHORS");
pub const VERSION: &str = env!("CARGO_PKG_VERSION");
pub const BUILDTD: &str = env!("BUILD_TIMEDATE");
pub const REVISION: &str = env!("VCS_COMMIT_HASH");
pub const REVDATE: &str = env!("VCS_COMMIT_DATE");

mod libs;

#[macro_use]
extern crate lazy_static;

use std::{collections::VecDeque, str::FromStr, time::Duration};

use crate::libs::*;
use a2s::A2SClient;
use chrono::SubsecRound;
use configuration::*;
use futures::StreamExt;
use serverlist::get_servers;
use serverlist::QueryServer;
use std::time::Instant;
use tokio::time::sleep_until;
use tracing::{debug, error, info, warn};
use tracing_log::log::trace;

#[derive(Debug)]
pub struct ServerResponse<'a> {
    pub server: &'a QueryServer,
    pub clients: u32,
    pub time: chrono::DateTime<chrono::offset::Utc>,
    pub map: Option<String>,
}

struct Conf {
    interval: f64,
    max_concurrent: u64,
}

fn main() -> Result<(), eyre::Error> {
    println!("╔════════════╤════════ SERVER TRACKER ════════════════════╗");
    println!("║ Version    │ {0: <43}║", VERSION);
    println!("║ Build Date │ {0: <43}║", BUILDTD);
    println!("║ Revision   │ {0: <43}║", REVISION);
    println!("║ Rev Date   │ {0: <43}║", REVDATE);
    println!("╚════════════╧════════════════════════════════════════════╝\n");
    let cfg = get_config_init_logging()?;
    info!("Starting");

    let servers = get_servers()?;
    info!(
        "Servers to query: [ {} ]",
        servers
            .iter()
            .map(|f| f.to_string())
            .collect::<Vec<String>>()
            .join(" ")
    );

    let conf = Conf {
        interval: cfg
            .getfloat("querying", "interval")
            .map_or(None, |f| f)
            .unwrap_or(10.0),
        max_concurrent: cfg
            .getuint("querying", "max-concurrent")
            .map_or(None, |f| f)
            .unwrap_or(5),
    };
    // Main loop
    tokio::runtime::Builder::new_multi_thread()
        .enable_all()
        .worker_threads(
            cfg.getuint("core", "worker-threads")
                .unwrap_or(Some(8))
                .unwrap()
                .min(128) as usize,
        )
        .build()
        .expect("Building Tokio runtime")
        .block_on(async { real_main(&conf, &servers, &cfg).await })?;
    unreachable!("Async loop should run forever.")
}

async fn real_main(
    conf: &Conf,
    servers: &Vec<QueryServer>,
    cfg: &configparser::ini::Ini,
) -> Result<(), eyre::Error> {
    let mut response_queue: VecDeque<ServerResponse> = VecDeque::new();
    let mut db_con: Option<influxdb::Client> = None;
    let mut next;
    let a2s = a2s::A2SClient::new()
        .await
        .expect("While creating A2S Client");
    loop {
        let time_now = chrono::Utc::now().round_subsecs(3);
        next = Instant::now() + Duration::from_secs_f64(conf.interval);

        let resp = query_servers(&servers, conf.max_concurrent, time_now, &a2s).await;
        debug!("Collected {} responses", resp.len());
        for sr in resp {
            response_queue.push_front(sr);
        }

        debug!("Writing to InfluxDB...");
        'connection: loop {
            match &db_con {
                None => match influx::connect(cfg).await {
                    Ok(v) => {
                        db_con = Some(v);
                        info!("Connected to InfluxDB");
                        continue 'connection;
                    }
                    Err(e) => {
                        error!("Couldn't connect to InfluxDB: {}", e);
                    }
                },
                Some(v) => {
                    let unwritten = futures::stream::iter(response_queue.drain(..))
                        .map(|data| async move {
                            match influx::write_playercount(v, &data).await {
                                Ok(v) => {
                                    trace!("Write Success: {:?}", v);
                                    None
                                }
                                Err(e) => {
                                    error!("Failed to write to InfluxDB: {}", e);
                                    Some(data)
                                }
                            }
                        })
                        .buffer_unordered((conf.max_concurrent as usize).max(1))
                        .collect::<Vec<Option<ServerResponse>>>()
                        .await
                        .into_iter()
                        .flatten()
                        .collect::<Vec<ServerResponse>>();

                    for stale in unwritten {
                        response_queue.push_back(stale);
                    }
                }
            }
            break 'connection;
        }
        sleep_until(next.into()).await;
    }
}

async fn query_servers<'a>(
    servers: &'a [QueryServer],
    max_concurrent: u64,
    time_now: chrono::DateTime<chrono::Utc>,
    client: &A2SClient,
) -> Vec<ServerResponse<'a>> {
    debug!("Querying {} servers.", servers.len());
    if servers.is_empty() {
        error!("Serverlist is empty.");
    }
    futures::stream::iter(&mut servers.iter().map(|server| {
        async move {
            match server.server_type {
                serverlist::ServerType::Source => match client.info(server.socket).await {
                    Ok(v) => {
                        trace!("Queried {}", server);
                        Some(ServerResponse {
                            server,
                            clients: v.players.into(),
                            time: time_now,
                            map: Some(v.map),
                        })
                    }
                    Err(e) => {
                        warn!("Couldn't query server: [ {} ] reason: {} ", server, e);
                        None
                    }
                },
                serverlist::ServerType::Minecraft => {
                    // connection config doesn't implement ToSocketAddr...
                    let config = async_minecraft_ping::ConnectionConfig::build(
                        server.socket.ip().to_string(),
                    )
                    .with_port(server.socket.port());
                    match config.connect().await {
                        Ok(v) => match v.status().await {
                            Ok(v) => {
                                trace!("Queried {}", server);
                                Some(ServerResponse {
                                    server,
                                    clients: v.status.players.online,
                                    time: time_now,
                                    map: None,
                                })
                            }
                            Err(e) => {
                                warn!("Couldn't query server: [ {} ] reason: {} ", server, e);
                                None
                            }
                        },
                        Err(e) => {
                            warn!("Couldn't query server: [ {} ] reason: {} ", server, e);
                            None
                        }
                    }
                }

                // Mumble server does not take from the connection pool,
                // instead we naively create new UDP socket per connection.
                // This is currently blocking, as writing a UDP socket pool,
                // is non-trivial.
                serverlist::ServerType::Mumble => match mumble::query_server(server, &time_now) {
                    Ok(v) => {
                        trace!("Queried {}", server);
                        Some(v)
                    }
                    Err(e) => {
                        warn!("Couldn't query server: [ {} ] reason: {} ", server, e);
                        None
                    }
                },
            }
        }
    }))
    .buffer_unordered((max_concurrent as usize).max(1))
    .collect::<Vec<Option<ServerResponse>>>()
    .await
    .into_iter()
    .flatten()
    .collect::<Vec<ServerResponse>>()
    // There's probably a more efficient way than collecting twice,
    // but I couldn't figure it out.
}

/// Initializes logging, and attemps to load the config file.
fn get_config_init_logging() -> Result<configparser::ini::Ini, eyre::Error> {
    use tracing_subscriber::filter::{EnvFilter, LevelFilter};

    // Logging and CFG
    enable_ansi()?;
    // TRACE level events will be output here.
    let builder = tracing_subscriber::fmt()
        .with_env_filter(EnvFilter::default().add_directive(LevelFilter::TRACE.into()))
        .with_filter_reloading();
    let filter_handle = builder.reload_handle();
    builder.init();
    // Load CFG, this should be done after logging is initialized.
    let cfg = get_settings()?;
    let mut new_filter = EnvFilter::default().add_directive(
        cfg.get("logging", "log_level")
            .unwrap_or_else(|| "tracing".into())
            .parse()?,
    );
    let custom_filter = cfg
        .get("logging", "custom_filter")
        .unwrap_or_else(|| "".into());
    if !custom_filter.is_empty() {
        match tracing_subscriber::filter::Directive::from_str(custom_filter.as_str()) {
            Ok(v) => new_filter = new_filter.add_directive(v),
            Err(e) => {
                error!("Unable to parse custom filter: '{}' - {}", custom_filter, e);
            }
        }
    }
    // Switch to user-defined logging level.
    filter_handle.modify(|layer| *layer = new_filter)?;
    Ok(cfg)
}

#[cfg(target_os = "windows")]
fn enable_ansi() -> Result<(), color_eyre::Report> {
    ansi_term::enable_ansi_support()
        .map_err(|e| eyre::Report::msg(format!("Cannot enable ANSI color on Windows: {}", e)))
}
#[cfg(not(target_os = "windows"))]
fn enable_ansi() -> Result<(), color_eyre::Report> {
    Ok(())
}
